package ru.redguy.extendedpistons;

import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.block.*;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.*;
import org.bukkit.inventory.DoubleChestInventory;
import org.bukkit.material.PistonBaseMaterial;

import java.util.Comparator;
import java.util.*;

public class WorldListener implements Listener {
    @EventHandler(priority = EventPriority.LOW, ignoreCancelled = true)
    public void onPistonChanged(BlockPistonExtendEvent e) {
        if (e instanceof BlockPistonExtendedExtendEvent) return;
        Block block = e.getBlock();
        BlockState state = block.getState();
        PistonBaseMaterial piston = (PistonBaseMaterial) state.getData();
        int maxBlocks = findNearbyPlayerMaxBlocks(block.getLocation());

        PistonExtendsChecker checker = new PistonExtendsChecker(e.getBlock().getWorld(), e.getBlock().getLocation(), piston.getFacing(), true, maxBlocks);

        if (!checker.a()) {
            e.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onBlockBreak(BlockBreakEvent e) {
        Material blockType = e.getBlock().getType();
        if (!(blockType.equals(Material.REDSTONE_BLOCK) || blockType.equals(Material.REDSTONE_TORCH_ON))) return;
        HashMap<BlockFace, BlockState> states = new HashMap<>();
        HashMap<BlockFace, Integer> powers = new HashMap<>();
        for (BlockFace face : new BlockFace[]{BlockFace.SELF, BlockFace.UP, BlockFace.DOWN, BlockFace.NORTH, BlockFace.SOUTH, BlockFace.EAST, BlockFace.WEST}) {
            states.put(face, e.getBlock().getRelative(face).getState());
            powers.put(face, e.getBlock().getRelative(face).getBlockPower());
        }
        Bukkit.getScheduler().runTask(ExtendedPistons.INSTANCE, () -> {
            if (ExtendedPistons.INSTANCE.isStickyPistonPulls()) {
                ArrayList<BlockFace> checkedBlockFaces = new ArrayList<BlockFace>() {{
                    add(BlockFace.SELF);
                    add(BlockFace.UP);
                    add(BlockFace.DOWN);
                }};
                if (e.getBlock().getType().equals(Material.REDSTONE_WIRE)) {
                    for (BlockFace blockFace : new BlockFace[]{BlockFace.NORTH, BlockFace.SOUTH, BlockFace.EAST, BlockFace.WEST}) {
                        Block block = e.getBlock().getRelative(blockFace);
                        if (block.getBlockPower() == 0 && powers.get(blockFace) != 0) checkedBlockFaces.add(blockFace);
                    }
                } else {
                    checkedBlockFaces.add(BlockFace.NORTH);
                    checkedBlockFaces.add(BlockFace.SOUTH);
                    checkedBlockFaces.add(BlockFace.EAST);
                    checkedBlockFaces.add(BlockFace.WEST);
                }
                for (BlockFace face : checkedBlockFaces) {
                    Block block = e.getBlock().getRelative(face);
                    BlockState state = states.get(face);
                    if (state.getType().equals(Material.PISTON_STICKY_BASE)) {
                        PistonBaseMaterial piston;
                        try {
                            piston = (PistonBaseMaterial) state.getData();
                        } catch (ClassCastException ex) {
                            continue;
                        }
                        int maxBlocks = findNearbyPlayerMaxBlocks(block.getLocation());

                        PistonExtendsChecker checker = new PistonExtendsChecker(e.getBlock().getWorld(), block.getLocation(), piston.getFacing(), false, maxBlocks);
                        boolean isMovable = checker.a();
                        List<Block> blocks = checker.getMovedBlocksObjects();
                        List<Block> toBreak = checker.getBrokenBlocksObjects();

                        if (isMovable) {
                            BlockPistonRetractEvent ev = new BlockPistonRetractEvent(block, blocks, piston.getFacing());
                            Bukkit.getPluginManager().callEvent(ev);
                            if (ev.isCancelled()) continue;
                            for (Block block1 : toBreak) {
                                block1.breakNaturally();
                            }
                            for (Block b : blocks) {
                                Block b2 = b.getRelative(piston.getFacing().getOppositeFace());
                                b2.setType(b.getType());
                                b2.setData(b.getData()); //Deprecated but powerful
                                if (b.getType().equals(Material.CHEST)) {
                                    Chest oldChest = (Chest) b.getState();
                                    Chest newChest = (Chest) b2.getState();
                                    newChest.getBlockInventory().setContents(oldChest.getBlockInventory().getContents());
                                    oldChest.getBlockInventory().clear();
                                } else if (b.getType().equals(Material.FURNACE) || b.getType().equals(Material.BURNING_FURNACE)) {
                                    Furnace oldFurnace = (Furnace) b.getState();
                                    Furnace newFurnace = (Furnace) b2.getState();
                                    newFurnace.getInventory().setContents(oldFurnace.getInventory().getContents());
                                    oldFurnace.getInventory().clear();
                                }
                                b.setType(Material.AIR);
                            }
                        }
                    }
                }
            }
        });
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onBlockPlace(BlockPlaceEvent e) {
        Material blockType = e.getBlock().getType();
        if (!(blockType.equals(Material.REDSTONE_BLOCK) || blockType.equals(Material.REDSTONE_TORCH_ON))) return;
        HashMap<BlockFace, BlockState> states = new HashMap<>();
        HashMap<BlockFace, Integer> powers = new HashMap<>();
        for (BlockFace face : new BlockFace[]{BlockFace.SELF, BlockFace.UP, BlockFace.DOWN, BlockFace.NORTH, BlockFace.SOUTH, BlockFace.EAST, BlockFace.WEST}) {
            states.put(face, e.getBlock().getRelative(face).getState());
            powers.put(face, e.getBlock().getRelative(face).getBlockPower());
        }
        Bukkit.getScheduler().runTask(ExtendedPistons.INSTANCE, () -> {
            ArrayList<BlockFace> checkedBlockFaces = new ArrayList<BlockFace>() {{
                add(BlockFace.SELF);
                add(BlockFace.UP);
                add(BlockFace.DOWN);
            }};
            if (e.getBlock().getType().equals(Material.REDSTONE_WIRE)) {
                for (BlockFace blockFace : new BlockFace[]{BlockFace.NORTH, BlockFace.SOUTH, BlockFace.EAST, BlockFace.WEST}) {
                    Block block = e.getBlock().getRelative(blockFace);
                    if (block.getBlockPower() != 0 && powers.get(blockFace) == 0) checkedBlockFaces.add(blockFace);
                }
            } else {
                checkedBlockFaces.add(BlockFace.NORTH);
                checkedBlockFaces.add(BlockFace.SOUTH);
                checkedBlockFaces.add(BlockFace.EAST);
                checkedBlockFaces.add(BlockFace.WEST);
            }
            for (BlockFace face : checkedBlockFaces) {
                Block block = e.getBlock().getRelative(face);
                if (block.getType().equals(Material.PISTON_BASE) || block.getType().equals(Material.PISTON_STICKY_BASE)) {
                    BlockState state = states.get(face);
                    PistonBaseMaterial piston;
                    try {
                        piston = (PistonBaseMaterial) state.getData();
                    } catch (ClassCastException ex) {
                        continue;
                    }
                    int maxBlocks = findNearbyPlayerMaxBlocks(block.getLocation());

                    PistonExtendsChecker checker = new PistonExtendsChecker(e.getBlock().getWorld(), block.getLocation(), piston.getFacing(), true, maxBlocks);
                    boolean isMovable = checker.a();
                    List<Block> blocks = checker.getMovedBlocksObjects();
                    List<Block> toBreak = checker.getBrokenBlocksObjects();

                    upper:
                    for (Block b : blocks) {
                        if(b.getType().equals(Material.CHEST)) {
                            Block b2 = b.getRelative(piston.getFacing());
                            ArrayList<BlockFace> faces = new ArrayList<BlockFace>() {{
                                add(BlockFace.NORTH);
                                add(BlockFace.SOUTH);
                                add(BlockFace.WEST);
                                add(BlockFace.EAST);
                            }};
                            faces.remove(piston.getFacing().getOppositeFace());
                            int chests = 0;
                            for (BlockFace blockFace : faces) {
                                Block b3 = b2.getRelative(blockFace);
                                if (b3.getType() == Material.CHEST) {
                                    chests++;
                                    Chest chest = (Chest) b3.getState();
                                    if (chest.getInventory() instanceof DoubleChestInventory) {
                                        isMovable = false;
                                        break upper;
                                    }
                                }
                            }
                            if (chests > 1) {
                                isMovable = false;
                                break;
                            } else if (chests == 1) {
                                if (b.getType() == Material.CHEST) {
                                    Chest chest = (Chest) b.getState();
                                    if (chest.getInventory() instanceof DoubleChestInventory) {
                                        isMovable = false;
                                        break;
                                    }
                                }
                            }
                        }
                    }

                    if (isMovable) {
                        BlockPistonExtendedExtendEvent event = new BlockPistonExtendedExtendEvent(block, blocks, piston.getFacing());
                        Bukkit.getPluginManager().callEvent(event);
                        if (event.isCancelled()) return;
                        Collections.reverse(blocks);
                        for (Block block1 : toBreak) {
                            block1.breakNaturally();
                        }
                        for (Block b : blocks) {
                            Block b2 = b.getRelative(piston.getFacing());
                            b2.setType(b.getType());
                            b2.setData(b.getData()); //Deprecated but powerful
                            if (b.getType().equals(Material.CHEST)) {
                                Chest oldChest = (Chest) b.getState();
                                Chest newChest = (Chest) b2.getState();
                                newChest.getBlockInventory().setContents(oldChest.getBlockInventory().getContents());
                                oldChest.getBlockInventory().clear();
                            } else if (b.getType().equals(Material.FURNACE) || b.getType().equals(Material.BURNING_FURNACE)) {
                                Furnace oldFurnace = (Furnace) b.getState();
                                Furnace newFurnace = (Furnace) b2.getState();
                                newFurnace.getInventory().setContents(oldFurnace.getInventory().getContents());
                                oldFurnace.getInventory().clear();
                            }
                            b.setType(Material.AIR);
                        }
                    }
                }
            }
        });
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onRedstone(BlockRedstoneEvent e) {
        HashMap<BlockFace, BlockState> states = new HashMap<>();
        for (BlockFace face : new BlockFace[]{BlockFace.SELF, BlockFace.UP, BlockFace.DOWN, BlockFace.NORTH, BlockFace.SOUTH, BlockFace.EAST, BlockFace.WEST}) {
            states.put(face, e.getBlock().getRelative(face).getState());
        }
        Bukkit.getScheduler().runTask(ExtendedPistons.INSTANCE, () -> {
            if (e.getOldCurrent() != 0 && e.getNewCurrent() == 0) {
                if (ExtendedPistons.INSTANCE.isStickyPistonPulls()) {
                    ArrayList<BlockFace> checkedBlockFaces = new ArrayList<BlockFace>() {{
                        add(BlockFace.SELF);
                        add(BlockFace.UP);
                        add(BlockFace.DOWN);
                    }};
                    if (e.getBlock().getType().equals(Material.REDSTONE_WIRE)) {
                        for (BlockFace blockFace : new BlockFace[]{BlockFace.NORTH, BlockFace.SOUTH, BlockFace.EAST, BlockFace.WEST}) {
                            Block block = e.getBlock().getRelative(blockFace);
                            if (block.getBlockPower() == 0) checkedBlockFaces.add(blockFace);
                        }
                    } else {
                        checkedBlockFaces.add(BlockFace.NORTH);
                        checkedBlockFaces.add(BlockFace.SOUTH);
                        checkedBlockFaces.add(BlockFace.EAST);
                        checkedBlockFaces.add(BlockFace.WEST);
                    }
                    for (BlockFace face : checkedBlockFaces) {
                        Block block = e.getBlock().getRelative(face);
                        BlockState state = states.get(face);
                        if (state.getType().equals(Material.PISTON_STICKY_BASE)) {
                            PistonBaseMaterial piston;
                            try {
                                piston = (PistonBaseMaterial) state.getData();
                            } catch (ClassCastException ex) {
                                continue;
                            }
                            int maxBlocks = findNearbyPlayerMaxBlocks(block.getLocation());

                            PistonExtendsChecker checker = new PistonExtendsChecker(e.getBlock().getWorld(), block.getLocation(), piston.getFacing(), false, maxBlocks);
                            boolean isMovable = checker.a();
                            List<Block> blocks = checker.getMovedBlocksObjects();
                            List<Block> toBreak = checker.getBrokenBlocksObjects();

                            if (isMovable) {
                                BlockPistonRetractEvent ev = new BlockPistonRetractEvent(block, blocks, piston.getFacing());
                                Bukkit.getPluginManager().callEvent(ev);
                                if (ev.isCancelled()) continue;
                                for (Block block1 : toBreak) {
                                    block1.breakNaturally();
                                }
                                for (Block b : blocks) {
                                    Block b2 = b.getRelative(piston.getFacing().getOppositeFace());
                                    b2.setType(b.getType());
                                    b2.setData(b.getData()); //Deprecated but powerful
                                    if (b.getType().equals(Material.CHEST)) {
                                        Chest oldChest = (Chest) b.getState();
                                        Chest newChest = (Chest) b2.getState();
                                        newChest.getBlockInventory().setContents(oldChest.getBlockInventory().getContents());
                                        oldChest.getBlockInventory().clear();
                                    } else if (b.getType().equals(Material.FURNACE) || b.getType().equals(Material.BURNING_FURNACE)) {
                                        Furnace oldFurnace = (Furnace) b.getState();
                                        Furnace newFurnace = (Furnace) b2.getState();
                                        newFurnace.getInventory().setContents(oldFurnace.getInventory().getContents());
                                        oldFurnace.getInventory().clear();
                                    }
                                    b.setType(Material.AIR);
                                }
                            }
                        }
                    }
                }
            } else {
                ArrayList<BlockFace> checkedBlockFaces = new ArrayList<BlockFace>() {{
                    add(BlockFace.SELF);
                    add(BlockFace.UP);
                    add(BlockFace.DOWN);
                }};
                if (e.getBlock().getType().equals(Material.REDSTONE_WIRE)) {
                    for (BlockFace blockFace : new BlockFace[]{BlockFace.NORTH, BlockFace.SOUTH, BlockFace.EAST, BlockFace.WEST}) {
                        Block block = e.getBlock().getRelative(blockFace);
                        if (block.getBlockPower() != 0) checkedBlockFaces.add(blockFace);
                    }
                } else {
                    checkedBlockFaces.add(BlockFace.NORTH);
                    checkedBlockFaces.add(BlockFace.SOUTH);
                    checkedBlockFaces.add(BlockFace.EAST);
                    checkedBlockFaces.add(BlockFace.WEST);
                }
                for (BlockFace face : checkedBlockFaces) {
                    Block block = e.getBlock().getRelative(face);
                    if (block.getType().equals(Material.PISTON_BASE) || block.getType().equals(Material.PISTON_STICKY_BASE)) {
                        BlockState state = states.get(face);
                        PistonBaseMaterial piston;
                        try {
                            piston = (PistonBaseMaterial) state.getData();
                        } catch (ClassCastException ex) {
                            continue;
                        }
                        int maxBlocks = findNearbyPlayerMaxBlocks(block.getLocation());

                        PistonExtendsChecker checker = new PistonExtendsChecker(e.getBlock().getWorld(), block.getLocation(), piston.getFacing(), true, maxBlocks);
                        boolean isMovable = checker.a();
                        List<Block> blocks = checker.getMovedBlocksObjects();
                        List<Block> toBreak = checker.getBrokenBlocksObjects();

                        upper:
                        for (Block b : blocks) {
                            if(b.getType().equals(Material.CHEST)) {
                                Block b2 = b.getRelative(piston.getFacing());
                                ArrayList<BlockFace> faces = new ArrayList<BlockFace>() {{
                                    add(BlockFace.NORTH);
                                    add(BlockFace.SOUTH);
                                    add(BlockFace.WEST);
                                    add(BlockFace.EAST);
                                }};
                                faces.remove(piston.getFacing().getOppositeFace());
                                int chests = 0;
                                for (BlockFace blockFace : faces) {
                                    Block b3 = b2.getRelative(blockFace);
                                    if (b3.getType() == Material.CHEST) {
                                        chests++;
                                        Chest chest = (Chest) b3.getState();
                                        if (chest.getInventory() instanceof DoubleChestInventory) {
                                            isMovable = false;
                                            break upper;
                                        }
                                    }
                                }
                                if (chests > 1) {
                                    isMovable = false;
                                    break;
                                } else if (chests == 1) {
                                    if (b.getType() == Material.CHEST) {
                                        Chest chest = (Chest) b.getState();
                                        if (chest.getInventory() instanceof DoubleChestInventory) {
                                            isMovable = false;
                                            break;
                                        }
                                    }
                                }
                            }
                        }

                        if (isMovable) {
                            BlockPistonExtendedExtendEvent event = new BlockPistonExtendedExtendEvent(block, blocks, piston.getFacing());
                            Bukkit.getPluginManager().callEvent(event);
                            if (event.isCancelled()) return;
                            Collections.reverse(blocks);
                            for (Block block1 : toBreak) {
                                block1.breakNaturally();
                            }
                            for (Block b : blocks) {
                                Block b2 = b.getRelative(piston.getFacing());
                                b2.setType(b.getType());
                                b2.setData(b.getData()); //Deprecated but powerful
                                if (b.getType().equals(Material.CHEST)) {
                                    Chest oldChest = (Chest) b.getState();
                                    Chest newChest = (Chest) b2.getState();
                                    newChest.getBlockInventory().setContents(oldChest.getBlockInventory().getContents());
                                    oldChest.getBlockInventory().clear();
                                } else if (b.getType().equals(Material.FURNACE) || b.getType().equals(Material.BURNING_FURNACE)) {
                                    Furnace oldFurnace = (Furnace) b.getState();
                                    Furnace newFurnace = (Furnace) b2.getState();
                                    newFurnace.getInventory().setContents(oldFurnace.getInventory().getContents());
                                    oldFurnace.getInventory().clear();
                                }
                                b.setType(Material.AIR);
                            }
                        }
                    }
                }
            }
        });
    }

    public int findNearbyPlayerMaxBlocks(Location location) {
        Optional<? extends Player> op = Bukkit.getOnlinePlayers().stream().min(Comparator.comparingDouble(p -> p.getLocation().distance(location)));
        if (op.isPresent()) {
            String[] groups = ExtendedPistons.INSTANCE.permsService.getPlayerGroups(op.get());
            Optional<Integer> intOp = Arrays.stream(groups).map(group -> (ExtendedPistons.INSTANCE.conf.isSet("piston-move-limitations." + group) && ExtendedPistons.INSTANCE.conf.isInt("piston-move-limitations." + group))?ExtendedPistons.INSTANCE.conf.getInt("piston-move-limitations." + group):0).max(Comparator.comparingInt(i -> i));
            if(intOp.isPresent()) {
                return intOp.get()==0?12:intOp.get();
            }
        }
        return 12;
    }
}
